#include <xkbcommon/xkbcommon.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <gio/gio.h>
#include "virtual-keyboard-unstable-v1.h"
#include "virtual-keyboard.h"
#include "wayland.h"

/**
 * @brief DiyaVkbButton: virtual keyboard button
 */

struct _DiyaVkbButton
{
    GtkWidget super;
    GtkWidget *labels[3];
};

G_DEFINE_FINAL_TYPE(DiyaVkbButton, diya_vkb_button, GTK_TYPE_WIDGET);

static void diya_vkb_button_dispose(GObject *object)
{
    DiyaVkbButton *self = DIYA_VKB_BUTTON(object);
    g_debug("diya_vkb_button_dispose");
    for (int i = 0; i < 3; i++)
    {
        g_clear_pointer(&self->labels[i], gtk_widget_unparent);
    }
    //  g_clear_pointer (&demo->label, gtk_widget_unparent);
    G_OBJECT_CLASS(diya_vkb_button_parent_class)->dispose(object);
}

static void diya_vkb_button_init(DiyaVkbButton *self)
{
    gchar class_name[32]; /*diya-vkb-btn-level-%d */
    GtkLayoutManager *grid_layout;
    grid_layout = gtk_widget_get_layout_manager(GTK_WIDGET(self));
    //gtk_orientable_set_orientation(GTK_ORIENTABLE(box_layout), GTK_ORIENTATION_VERTICAL);
    // gtk_box_layout_set_spacing (GTK_BOX_LAYOUT (box_layout), 3);
    gtk_widget_add_css_class(GTK_WIDGET(self),"diya-vkb-btn");
    gtk_widget_set_can_focus(GTK_WIDGET(self), false);
    // init labels
    for (int i = 0; i < 3; i++)
    {
        self->labels[i] = gtk_label_new(i==0?"X":"x");
        g_snprintf(class_name, sizeof(class_name),"diya-vkb-btn-level-%d",i + 1);
        gtk_widget_add_css_class(self->labels[i], class_name);
        gtk_widget_set_hexpand(self->labels[i], true);

        gtk_widget_set_parent(self->labels[i], GTK_WIDGET(self));

        GtkGridLayoutChild *grid_child = GTK_GRID_LAYOUT_CHILD(gtk_layout_manager_get_layout_child(grid_layout,self->labels[i]));
        switch (i)
        {
        case 0:
            /* code */
            gtk_grid_layout_child_set_row(grid_child,0);
            gtk_grid_layout_child_set_column(grid_child,0);
            gtk_grid_layout_child_set_column_span (grid_child,2);
            break;
        case 1:
            gtk_grid_layout_child_set_row(grid_child,1);
            gtk_grid_layout_child_set_column(grid_child,0);
            gtk_widget_set_halign(self->labels[i],GTK_ALIGN_START);
            break;
        case 2:
            gtk_grid_layout_child_set_row(grid_child,1);
            gtk_grid_layout_child_set_column(grid_child,1);
            gtk_widget_set_halign(self->labels[i],GTK_ALIGN_END);
            break;
        default:
            break;
        }
    }
}


static void diya_vkb_button_class_init(DiyaVkbButtonClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->dispose = diya_vkb_button_dispose;

    gtk_widget_class_set_layout_manager_type(GTK_WIDGET_CLASS(class), GTK_TYPE_GRID_LAYOUT);
}

DiyaVkbButton* diya_vkb_button_new(/**with prameter */)
{
    return DIYA_VKB_BUTTON(g_object_new(DIYA_TYPE_VKB_BUTTON, NULL));
}


/**
 * @brief DiyaVirtualKeyboardWidget: Virtual keyboard GTK widget
 *
 */

#define VKB_N_CELLS 145
#define VKB_ROW_NUM(i) (i/29)
#define VKB_COL_NUM(i) (i % 29)

struct _DiyaVirtualKeyboardWidget
{
    GtkWidget super;
    /**
     * row 0: 00 - 13
     * row 1: 14 - 27
     * row 2: 28 - 40
     * row 3: 41 - 53
     * row 4: 54 - 62
     */
    GList *buttons;
};

G_DEFINE_FINAL_TYPE(DiyaVirtualKeyboardWidget, diya_virtual_keyboard_widget, GTK_TYPE_WIDGET);

static void diya_virtual_keyboard_widget_dispose(GObject *object)
{
    DiyaVirtualKeyboardWidget *self = DIYA_VIRTUAL_KEYBOARD_WIDGET(object);
    g_debug("diya_virtual_keyboard_widget_dispose");
    if(self->buttons)
    {
        GList * it = NULL;
        for(it = self->buttons; it; it = it->next)
        {
            if(it->data)
            {
                g_clear_pointer(&it->data, gtk_widget_unparent);
            }
        }
    }
    //  g_clear_pointer (&demo->label, gtk_widget_unparent);
    G_OBJECT_CLASS(diya_virtual_keyboard_widget_parent_class)->dispose(object);
}

static void diya_virtual_keyboard_widget_init(DiyaVirtualKeyboardWidget *self)
{
    // gtk_widget_set_size_request(GTK_WIDGET(self), 600, 400);
    g_debug("diya_virtual_keyboard_widget_init: Create new widget");
    GtkLayoutManager *grid_layout;
    gtk_widget_add_css_class(GTK_WIDGET(self),"diya-vkb");
    grid_layout = gtk_widget_get_layout_manager(GTK_WIDGET(self));
    // gtk_orientable_set_orientation(GTK_ORIENTABLE(grid_layout), GTK_ORIENTATION_VERTICAL);
    gtk_grid_layout_set_row_spacing(GTK_GRID_LAYOUT(grid_layout), 2);
    gtk_grid_layout_set_column_spacing(GTK_GRID_LAYOUT(grid_layout), 2);
    //gtk_grid_layout_set_baseline_row(GTK_GRID_LAYOUT(grid_layout), 4);
    // init buttons
    for (int i = 0; i < VKB_N_CELLS; i++)
    {
        gboolean skip = false;
        int row = VKB_ROW_NUM(i);
        int col = VKB_COL_NUM(i);
        switch (row)
        {
        case 0:
            skip = (col %2 != 0) || (col > 26);
            break;
        case 1:
            skip = (col == 1) || (col == 2) || (col != 0 && col % 2 == 0);
            break;
        case 2:
            skip = (col == 1) || (col == 2) || (col != 0 && col%2 == 0) || (col > 25);
            break;
        case 3:
            skip = (col > 0 && col < 5) || (col != 0 && col % 2 == 0);
            break;
        case 4:
            skip = (col > 0 && col < 3) || (col == 4) || (col==6) || (col>7 && col < 19) || (col > 19 && col %2 == 0); 
            break;
        default:
            break;
        }

        if(skip)
        {
            continue;
        }
        
        GtkWidget* button = GTK_WIDGET(diya_vkb_button_new());
        self->buttons = g_list_append(self->buttons, button);
        gtk_widget_set_parent(button, GTK_WIDGET(self));
        gtk_widget_set_hexpand(button, true);
        GtkGridLayoutChild *grid_child = GTK_GRID_LAYOUT_CHILD(gtk_layout_manager_get_layout_child(grid_layout,button));
        
        gtk_grid_layout_child_set_row(grid_child,row);
        gtk_grid_layout_child_set_column(grid_child,col);
    
        switch (i)
        {
        case 26:
        case 29:
        case 58:
        case 116:
            gtk_grid_layout_child_set_column_span (grid_child,3);
            break;
        
        case 83:
            gtk_grid_layout_child_set_column_span (grid_child,4);
            break;
        case 87:
            gtk_grid_layout_child_set_column_span (grid_child,5);
            break;
        case 123:
            gtk_grid_layout_child_set_column_span (grid_child,12);
            break;
        
        default:
            if(i > 0)
                gtk_grid_layout_child_set_column_span (grid_child,2);
            break;
        }
    }
}

static void diya_virtual_keyboard_widget_class_init(DiyaVirtualKeyboardWidgetClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->dispose = diya_virtual_keyboard_widget_dispose;

    gtk_widget_class_set_layout_manager_type(GTK_WIDGET_CLASS(class), GTK_TYPE_GRID_LAYOUT);
}

DiyaVirtualKeyboardWidget *diya_virtual_keyboard_widget_new()
{
    DiyaVirtualKeyboardWidget *vkb_widget = g_object_new(DIYA_TYPE_VIRTUAL_KEYBOARD_WIDGET, NULL);
    return vkb_widget;
}

/**
 * @brief  DiyaVirtualKeyboard class: lowlevel wayland virtual keyboard handle
 *
 */

static struct zwp_virtual_keyboard_manager_v1 *g_virtual_keyboard_manager = NULL;

struct _DiyaVirtualKeyboard
{
    DiyaShellObject parent;
    struct zwp_virtual_keyboard_v1 *keyboard;
};

G_DEFINE_FINAL_TYPE(DiyaVirtualKeyboard, diya_virtual_keyboard, DIYA_TYPE_SHELL_OBJECT);

static void diya_virtual_keyboard_dispose(GObject *object)
{
    DiyaVirtualKeyboard *self = DIYA_VIRTUAL_KEYBOARD(object);
    g_debug("diya_virtual_keyboard_dispose");
    if (self->keyboard)
    {
        zwp_virtual_keyboard_v1_destroy(self->keyboard);
    }
    G_OBJECT_CLASS(diya_virtual_keyboard_parent_class)->dispose(object);
}

static void diya_virtual_keyboard_init(DiyaVirtualKeyboard *self)
{
    self->keyboard = NULL;
}

static const gchar *diya_virtual_keyboard_to_string(DiyaObject *object)
{
    (void)object;
    return "Diya virtual keyboard instance";
}

static void diya_virtual_keyboard_class_init(DiyaVirtualKeyboardClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);

    gobject_class->dispose = diya_virtual_keyboard_dispose;
    // gobject_class->set_property = diya_lock_session_set_property;
    // gobject_class->get_property = diya_lock_session_get_property;
    base_class->to_string = diya_virtual_keyboard_to_string;
}

/*
static void randname(char *buf)
{
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    long r = ts.tv_nsec;
    for (int i = 0; i < 6; ++i)
    {
        buf[i] = 'A' + (r & 15) + (r & 16) * 2;
        r >>= 5;
    }
}
*/

static int create_keymap_fd(off_t size)
{
    static const char name[] = "/diya-shell-keymap-XXXXXX";
    const char *base;
    char *path;
    int fd;
    int ret;
    // randname(name + sizeof(name) - 7);
    base = getenv("XDG_RUNTIME_DIR");
    if (!base)
    {
        errno = ENOENT;
        return -1;
    }
    path = malloc(strlen(base) + sizeof(name) + 1);
    if (!path)
        return -1;

    strcpy(path, base);
    strcat(path, name);
    g_debug("Create temp file for keymap: %s", path);

    fd = mkstemp(path);
    if (fd >= 0)
    {
        long flags;
        flags = fcntl(fd, F_GETFD);
        if (flags == -1)
        {
            g_critical("fcntl Unable to F_GETFD: %s", strerror(errno));
            close(fd);
            fd = -1;
        }
        else if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
        {
            g_critical("fcntl Unable to F_SETFD(FD_CLOEXEC): %s", strerror(errno));
            close(fd);
            fd = -1;
        }
        unlink(path);
    }
    else
    {
        g_critical("mkstemp Unable to create temp file %s: %s", path, strerror(errno));
    }
    free(path);

    if (fd < 0)
        return -1;
    do
    {
        ret = ftruncate(fd, size);
    } while (ret < 0 && errno == EINTR);
    if (ret < 0)
    {
        close(fd);
        return -1;
    }

    return fd;
}

DiyaVirtualKeyboard *diya_virtual_keyboard_new(DiyaShell *shell, const gchar *keymap_file)
{
    gchar *content = NULL;
    GError *error = NULL;
    GBytes *bytes = NULL;
    gsize len = 0;

    if (keymap_file)
    {
        g_file_get_contents(keymap_file, &content, &len, &error);
        if (error != NULL)
        {
            g_critical("diya_virtual_keyboard_new: Unable to read file %s: %s", keymap_file, error->message);
            g_error_free(error);
            return NULL;
        }
    }
    else
    {
        g_warning("No keymap file specified. Loading default keymap from resource");
        GBytes *bytes = g_resources_lookup_data("/dev/iohub/diya/shell/default.keymap", 0, &error);
        if (error != NULL)
        {
            g_critical("Unable to read keymap file from resource %s", error->message);
            g_error_free(error);
            return NULL;
        }
        content = (gchar *)g_bytes_get_data(bytes, &len);
    }
    g_debug("Using default keymap configuration: \n%s", content);
    int fd = create_keymap_fd(len);
    if (fd == -1)
    {
        g_critical("diya_virtual_keyboard_new: create temp file");
        if (bytes)
        {
            g_bytes_unref(bytes);
            content = NULL;
        }
        if (content)
        {
            g_free(content);
        }
    }

    void *ptr = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (ptr == (void *)-1)
    {
        g_critical("diya_virtual_keyboard_new: error mmap: %s", strerror(errno));
        close(fd);
        if (bytes)
        {
            g_bytes_unref(bytes);
            content = NULL;
        }
        if (content)
        {
            g_free(content);
        }
    }

    strcpy((char *)ptr, content);

    DiyaVirtualKeyboard *vkb = g_object_new(DIYA_TYPE_VIRTUAL_KEYBOARD, "shell", shell, NULL);
    DiyaWayland *wayland = diya_shell_get_wayland(shell);
    struct wl_seat *seat = diya_wayland_get_seat(wayland);
    vkb->keyboard = zwp_virtual_keyboard_manager_v1_create_virtual_keyboard(g_virtual_keyboard_manager, seat);
    zwp_virtual_keyboard_v1_keymap(vkb->keyboard, WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1, fd, len);

    return vkb;
}

void diya_virtual_keyboard_register(struct wl_registry *registry, uint32_t name, DiyaShell *shell)
{
    (void)shell;
    g_virtual_keyboard_manager = wl_registry_bind(registry, name, &zwp_virtual_keyboard_manager_v1_interface, 1);
}

void diya_virtual_keyboard_send_key(DiyaVirtualKeyboard *self, uint32_t key, diya_vkb_state_t state)
{
    GDateTime *now = g_date_time_new_now_local();
    uint32_t current_time_ms = g_date_time_get_microsecond(now) / 1000;
    g_date_time_unref(now);
    zwp_virtual_keyboard_v1_key(self->keyboard, current_time_ms, key, (uint32_t)state);
}